/* USER CODE BEGIN Header */
/**
  ******************************************************************************
  * @file           : main.c
  * @brief          : Main program body
  ******************************************************************************
  ** This notice applies to any and all portions of this file
  * that are not between comment pairs USER CODE BEGIN and
  * USER CODE END. Other portions of this file, whether 
  * inserted by the user or by software development tools
  * are owned by their respective copyright owners.
  *
  * COPYRIGHT(c) 2020 STMicroelectronics
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
  *   1. Redistributions of source code must retain the above copyright notice,
  *      this list of conditions and the following disclaimer.
  *   2. Redistributions in binary form must reproduce the above copyright notice,
  *      this list of conditions and the following disclaimer in the documentation
  *      and/or other materials provided with the distribution.
  *   3. Neither the name of STMicroelectronics nor the names of its contributors
  *      may be used to endorse or promote products derived from this software
  *      without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************
  */
/* USER CODE END Header */

/* Includes ------------------------------------------------------------------*/
#include "main.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include "si5351.h"
/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */

/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */

/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
ADC_HandleTypeDef hadc1;

I2C_HandleTypeDef hi2c1;

/* USER CODE BEGIN PV */
/* Private variables ---------------------------------------------------------*/

/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_ADC1_Init(void);
static void MX_I2C1_Init(void);
/* USER CODE BEGIN PFP */
/* Private function prototypes -----------------------------------------------*/

/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */

#define CH_CW  0
#define CH_VFO 2
#define CH_BFO 1

// note: can be affected a little by the actual load in the circuit
const int32_t XtalFilterCenterFrequency = 8998300;
const int32_t si5351_correction = 5277;
int32_t Fvfo = 1000000; // dummy value, will be changed below

const si5351DriveStrength_t VFODriveStrength = SI5351_DRIVE_STRENGTH_4MA;
const si5351DriveStrength_t BFODriveStrength = SI5351_DRIVE_STRENGTH_4MA;
const si5351DriveStrength_t CWDriveStrength = SI5351_DRIVE_STRENGTH_8MA;
const int32_t qskDelayMs = 900;

int32_t targetFrequency = 10116000; // will be changed below, if necessary

const int32_t freqs[] = {
	10113500,
	10116000,
	10117500,
};

int32_t sendStart = 0;
int32_t sendFinish = 0;
int32_t keyReadTimeout = 0;
int32_t sendTimeout = 0;
int32_t keyerSpeedWPM = 19;
// keyerDitTimeMs = 60*1000/(50*WPM)
// where 50 is the length in dits of "PARIS "
// see https://morsecode.world/international/timing.html
int32_t keyerDitTimeMs = 63; // 19 WPM
bool straightKeyer = true;
bool isSending = false;

typedef enum {
	CW_SEND_NONE = 0,
	CW_SEND_DIT = 1,
	CW_SEND_DAH = 2,
} CwSend_t;

CwSend_t lastSent = CW_SEND_NONE;
CwSend_t nextSend = CW_SEND_NONE;

bool inTransmitMode = false;
uint32_t transmitModeEnterTime = 0;

void SetupCLK(uint8_t output, int32_t Fclk, si5351DriveStrength_t driveStrength) {
	static bool pllSetupDone = false;
	si5351PLLConfig_t pll_conf;
	si5351OutputConfig_t out_conf;

	si5351_Calc(Fclk, &pll_conf, &out_conf);
	if(!pllSetupDone) {
		si5351_SetupPLL(SI5351_PLL_A, &pll_conf);
		pllSetupDone = true;
	}
	si5351_SetupOutput(output, SI5351_PLL_A, driveStrength, &out_conf, 0);
}


void changeKeyerSpeed(int32_t speed) {
	keyerSpeedWPM = speed;
	if(keyerSpeedWPM < 10) {
		keyerSpeedWPM = 9;
		straightKeyer = true;
	} else {
		straightKeyer = false;
		if(keyerSpeedWPM > 35) {
			keyerSpeedWPM = 35;
		}
	}

	keyerDitTimeMs = 60*1000/(50*keyerSpeedWPM);
}

void changeFrequency(int32_t freq) {
	targetFrequency = freq;

	// use lowest VFO possible; experiments show that
	// with Si5351 this creates less spurious tones (birdies)
	if(targetFrequency > XtalFilterCenterFrequency) {
		Fvfo = targetFrequency - XtalFilterCenterFrequency;
	} else {
		Fvfo = XtalFilterCenterFrequency - targetFrequency;
	}

	si5351_EnableOutputs((1 << CH_BFO));
	SetupCLK(CH_VFO, Fvfo, VFODriveStrength);
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO));
}

void keyDown() {
	HAL_GPIO_WritePin(GPIOA, GPIO_PIN_8, GPIO_PIN_SET);
}

void keyUp() {
	HAL_GPIO_WritePin(GPIOA, GPIO_PIN_8, GPIO_PIN_RESET);
}

void switchRelayTx() {
	HAL_GPIO_WritePin(GPIOB, GPIO_PIN_14, GPIO_PIN_SET);
}

void switchRelayRx() {
	HAL_GPIO_WritePin(GPIOB, GPIO_PIN_14, GPIO_PIN_RESET);
}

void ensureTransmitMode() {
	if(inTransmitMode) {
		return;
	}

	si5351_EnableOutputs(0);
	keyUp();

	// attenuate our own received signal by shifting VFO/BFO frequencies
	int32_t delta = 10000;
	if(targetFrequency > XtalFilterCenterFrequency) {
		SetupCLK(CH_VFO, Fvfo-delta, VFODriveStrength);
	} else {
		SetupCLK(CH_VFO, Fvfo+delta, VFODriveStrength);
	}
	SetupCLK(CH_BFO, XtalFilterCenterFrequency + 1000 + delta, BFODriveStrength);
	SetupCLK(CH_CW, targetFrequency, CWDriveStrength);
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO)|(1 << CH_CW));

	switchRelayTx();
	inTransmitMode = true;
}

void ensureReceiveMode() {
	if(!inTransmitMode) {
		return;
	}

	// Restore the original VFO and BFO
	si5351_EnableOutputs(0);
	SetupCLK(CH_VFO, Fvfo, VFODriveStrength);
	SetupCLK(CH_BFO, XtalFilterCenterFrequency + 1000, BFODriveStrength);
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO));

	switchRelayRx();
	inTransmitMode = false;
}

void init() {
	// turn off the green LED on Blue Pill
	HAL_GPIO_WritePin(GPIOC, GPIO_PIN_13, GPIO_PIN_SET);
	
	/*
	// Code for determining the correction factor for Si5351
	si5351_Init(0);
	si5351PLLConfig_t pll_conf;
	si5351OutputConfig_t out_conf;
	int32_t Fclk = 10000000;

	si5351_Calc(Fclk, &pll_conf, &out_conf);
	si5351_SetupPLL(SI5351_PLL_A, &pll_conf);
	si5351_SetupOutput(CH_BFO, SI5351_PLL_A, BFODriveStrength, &out_conf, 0);
	si5351_EnableOutputs(1<<CH_BFO);

	while(1) {
	    HAL_GPIO_TogglePin(GPIOC, GPIO_PIN_13);
        HAL_Delay(1000);
	}
	*/

	keyUp();
	switchRelayRx();

	HAL_ADC_Start(&hadc1);

	si5351_Init(si5351_correction);
	SetupCLK(CH_BFO, XtalFilterCenterFrequency + 1000, BFODriveStrength);
	changeFrequency(freqs[0]);
}

bool buttonDitPressed() {
	return (HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_12) == GPIO_PIN_RESET);
}

bool buttonDahPressed() {
	return (HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_13) == GPIO_PIN_RESET);
}

bool checkKeyerEnabled() {
	return (HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_15) == GPIO_PIN_RESET);
}

void processStraightKeyerLogic(bool pressed) {
	if(pressed) {
		keyDown();
	} else {
		keyUp();
	}
}

void processIambicKeyerLogic(bool ditPressed, bool dahPressed) {
	int32_t now = HAL_GetTick();

	if(isSending && (now >= sendFinish)) {
		keyUp();
		isSending = false;
	}

	if(nextSend == CW_SEND_NONE) {
		if(ditPressed && (lastSent == CW_SEND_DAH)) {
			nextSend = CW_SEND_DIT;
		} else if(dahPressed && (lastSent == CW_SEND_DIT)) {
			nextSend = CW_SEND_DAH;
		} else if(now > keyReadTimeout) {
			if(ditPressed) {
				nextSend = CW_SEND_DIT;
			} else if (dahPressed) {
				nextSend = CW_SEND_DAH;
			}
		}
	}

	if((now > sendTimeout) && (nextSend != CW_SEND_NONE)) {
		if(nextSend == CW_SEND_DIT) {
			sendFinish = now + keyerDitTimeMs;
			keyReadTimeout = now + keyerDitTimeMs*2;
		} else if(nextSend == CW_SEND_DAH) {
			sendFinish = now + keyerDitTimeMs*3;
			keyReadTimeout = now + keyerDitTimeMs*3;
		}

		lastSent = nextSend;
		nextSend = CW_SEND_NONE;
		sendStart = now;
		sendTimeout = sendFinish + keyerDitTimeMs;
		isSending = true;
		keyDown();
	}
}

void loop() {
	int32_t adcVal;
	bool ditPressed = buttonDitPressed();
	bool dahPressed = buttonDahPressed();
	if(ditPressed || dahPressed) {
		transmitModeEnterTime = HAL_GetTick();
		ensureTransmitMode();
	} else {
		uint32_t tstamp = HAL_GetTick();
		if(tstamp - transmitModeEnterTime > qskDelayMs) {
			ensureReceiveMode();
		}
	}

	if(inTransmitMode) {
		if(straightKeyer) {
			processStraightKeyerLogic(ditPressed);
		} else {
			processIambicKeyerLogic(ditPressed, dahPressed);
		}
	} else {
		if(checkKeyerEnabled()) {
			if(straightKeyer) {
				changeKeyerSpeed(19);
			}
		} else {
			if(!straightKeyer) {
				changeKeyerSpeed(0);
			}
		}

		HAL_ADC_PollForConversion(&hadc1, HAL_MAX_DELAY);
        adcVal = HAL_ADC_GetValue(&hadc1);

        if(adcVal < 1000) {
        	if(targetFrequency != freqs[0]) {
        		changeFrequency(freqs[0]);
        	}
        } else if(adcVal > 3000) {
        	if(targetFrequency != freqs[2]) {
        		changeFrequency(freqs[2]);
        	}
        } else {
        	if(targetFrequency != freqs[1]) {
        		changeFrequency(freqs[1]);
        	}
        }
	}

    HAL_Delay(5);
}

/* USER CODE END 0 */

/**
  * @brief  The application entry point.
  * @retval int
  */
int main(void)
{
  /* USER CODE BEGIN 1 */

  /* USER CODE END 1 */

  /* MCU Configuration--------------------------------------------------------*/

  /* Reset of all peripherals, Initializes the Flash interface and the Systick. */
  HAL_Init();

  /* USER CODE BEGIN Init */

  /* USER CODE END Init */

  /* Configure the system clock */
  SystemClock_Config();

  /* USER CODE BEGIN SysInit */

  /* USER CODE END SysInit */

  /* Initialize all configured peripherals */
  MX_GPIO_Init();
  MX_ADC1_Init();
  MX_I2C1_Init();
  /* USER CODE BEGIN 2 */
  init();
  /* USER CODE END 2 */

  /* Infinite loop */
  /* USER CODE BEGIN WHILE */
  while (1)
  {
    loop();
    /* USER CODE END WHILE */

    /* USER CODE BEGIN 3 */

  }
  /* USER CODE END 3 */
}

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};
  RCC_PeriphCLKInitTypeDef PeriphClkInit = {0};

  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_NONE;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }
  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_HSI;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_0) != HAL_OK)
  {
    Error_Handler();
  }
  PeriphClkInit.PeriphClockSelection = RCC_PERIPHCLK_ADC;
  PeriphClkInit.AdcClockSelection = RCC_ADCPCLK2_DIV2;
  if (HAL_RCCEx_PeriphCLKConfig(&PeriphClkInit) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief ADC1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_ADC1_Init(void)
{

  /* USER CODE BEGIN ADC1_Init 0 */

  /* USER CODE END ADC1_Init 0 */

  ADC_ChannelConfTypeDef sConfig = {0};

  /* USER CODE BEGIN ADC1_Init 1 */

  /* USER CODE END ADC1_Init 1 */
  /**Common config 
  */
  hadc1.Instance = ADC1;
  hadc1.Init.ScanConvMode = ADC_SCAN_DISABLE;
  hadc1.Init.ContinuousConvMode = ENABLE;
  hadc1.Init.DiscontinuousConvMode = DISABLE;
  hadc1.Init.ExternalTrigConv = ADC_SOFTWARE_START;
  hadc1.Init.DataAlign = ADC_DATAALIGN_RIGHT;
  hadc1.Init.NbrOfConversion = 1;
  if (HAL_ADC_Init(&hadc1) != HAL_OK)
  {
    Error_Handler();
  }
  /**Configure Regular Channel 
  */
  sConfig.Channel = ADC_CHANNEL_1;
  sConfig.Rank = ADC_REGULAR_RANK_1;
  sConfig.SamplingTime = ADC_SAMPLETIME_1CYCLE_5;
  if (HAL_ADC_ConfigChannel(&hadc1, &sConfig) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN ADC1_Init 2 */

  /* USER CODE END ADC1_Init 2 */

}

/**
  * @brief I2C1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_I2C1_Init(void)
{

  /* USER CODE BEGIN I2C1_Init 0 */

  /* USER CODE END I2C1_Init 0 */

  /* USER CODE BEGIN I2C1_Init 1 */

  /* USER CODE END I2C1_Init 1 */
  hi2c1.Instance = I2C1;
  hi2c1.Init.ClockSpeed = 400000;
  hi2c1.Init.DutyCycle = I2C_DUTYCYCLE_2;
  hi2c1.Init.OwnAddress1 = 0;
  hi2c1.Init.AddressingMode = I2C_ADDRESSINGMODE_7BIT;
  hi2c1.Init.DualAddressMode = I2C_DUALADDRESS_DISABLE;
  hi2c1.Init.OwnAddress2 = 0;
  hi2c1.Init.GeneralCallMode = I2C_GENERALCALL_DISABLE;
  hi2c1.Init.NoStretchMode = I2C_NOSTRETCH_DISABLE;
  if (HAL_I2C_Init(&hi2c1) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN I2C1_Init 2 */

  /* USER CODE END I2C1_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOC_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOB_CLK_ENABLE();

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOC, GPIO_PIN_13, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_14, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_8, GPIO_PIN_RESET);

  /*Configure GPIO pin : PC13 */
  GPIO_InitStruct.Pin = GPIO_PIN_13;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);

  /*Configure GPIO pins : PB12 PB13 PB15 */
  GPIO_InitStruct.Pin = GPIO_PIN_12|GPIO_PIN_13|GPIO_PIN_15;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /*Configure GPIO pin : PB14 */
  GPIO_InitStruct.Pin = GPIO_PIN_14;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /*Configure GPIO pin : PA8 */
  GPIO_InitStruct.Pin = GPIO_PIN_8;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

}

/* USER CODE BEGIN 4 */

/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */
  while(1) 
  {
  }
  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{ 
  /* USER CODE BEGIN 6 */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
